import logging

from circuits import Component, handler

from brewpi_service.database import db_session, get_or_create
from brewpi_service.controller.models import Controller

from flask import Flask

app = Flask(__name__)
app.config["REDIS_URL"] = "redis://localhost"

from flask_sse import sse

from ..abstract import AbstractForwarder


LOGGER = logging.getLogger(__name__)


class SSEForwarder(Component, AbstractForwarder):
    """
    Notify SSE clients
    """
    @handler("ControllerConnected")
    def on_controller_appeared(self, event):
        LOGGER.debug("SSE: Controller Connected...")
        with app.app_context():
            sse.publish({"message": "connect"}, type='greeting')


    @handler("ControllerDisconnected")
    def on_controller_disappeared(self, event):
        LOGGER.debug("SSE: Controller disconnected...")
        with app.app_context():
            sse.publish({"message": "disconnect"}, type='greeting')




from sqlalchemy import (
    Column, Integer, String, Float, ForeignKey
)
from sqlalchemy.orm import relationship, backref


from brewpi_service.controller.models import ControllerBlock, BaseObject

from brewpi_service.controller.state import ControllerDataField

from zope.interface import implementer

from .interfaces import (
    ISetpoint,
    ISensorSetpointPair,
    ISensor
)

class ProcessValueMixin(BaseObject):
    __tablename__ = 'process_value_mixin'

    __mapper_args_ = {
        'polymorphic_identity': 'process_value_mixin'
    }


@implementer(ISetpoint)
class SetpointSimple(ControllerBlock):
    """
    A Simple Setpoint
    """
    __tablename__ = 'controller_block_setpoint_simple'

    __mapper_args__ = {
        'polymorphic_identity': "controller_block_setpoint_simple"
    }

    simple_setpoint_id = Column(Integer, ForeignKey('controller_block.id', ondelete='CASCADE'), primary_key=True)

    name = Column(String, nullable=True)

    value = ControllerDataField(Float)

    def __repr__(self):
        return "<SetpointSimple '{0}':{1}>".format(self.name,
                                                  self.value)


@implementer(ISensorSetpointPair)
class SensorSetpointPair(ControllerBlock):
    """
    A Simple pair of a sensor and a setpoint
    """
    __tablename__ = 'controller_block_sensor_setpoint_pair'

    __mapper_args__ = {
        'polymorphic_identity': "controller_block_sensor_setpoint_pair"
    }

    # sensor_setpoint_pair_id = Column(Integer, ForeignKey('controller_block.id'), primary_key=True)

    setpoint_id = Column(Integer, ForeignKey('controller_block.id'), nullable=True)
    setpoint = relationship("ControllerBlock", primaryjoin="and_(SensorSetpointPair.setpoint_id==ControllerBlock.id)", backref="setpoint_setpoint_pair")

    # sensor_id = Column(Integer, ForeignKey('controller_block.id'))
    # sensor = relationship("ControllerBlock", primaryjoin="and_(SensorSetpointPair.sensor_id==ControllerBlock.id)", backref="sensor_setpoint_pair")

    def __repr__(self):
        return '<SetPointPair set:{0}>'.format(self.setpoint)



@implementer(ISensor)
class TemperatureSensor(ControllerBlock):
    """
    A DS2xxx Temperature Sensor
    """
    __tablename__ = 'controller_block_ds2xxx_sensor'

    __mapper_args__ = {
        'polymorphic_identity': "controller_block_ds2xxx_sensor"
    }
    ds2xxx_sensor_id = Column(Integer, ForeignKey('controller_block.id', ondelete="CASCADE"), primary_key=True)

    value = ControllerDataField(Float)

    def __repr__(self):
        return "<Temperature Sensor '{0} at {1}°C'>".format(self.name, self.value)


class PID(ControllerBlock):
    """
    A proportional–integral–derivative controller (PID controller) is a control
    loop feedback mechanism (controller) commonly used in industrial control
    systems.
    """
    __tablename__ = 'controller_block_pid'

    id = Column(Integer, ForeignKey('controller_block.id', ondelete='CASCADE'), primary_key=True)

    __mapper_args__ = {
        'polymorphic_identity': "controller_block_pid",
        'inherit_condition': ControllerBlock.id == id
    }


    setpoint_id = Column(Integer, ForeignKey(ControllerBlock.id), nullable=True)
    setpoint = relationship('ControllerBlock',
                            primaryjoin="ControllerBlock.id == PID.setpoint_id",
                            foreign_keys=[setpoint_id],
                            backref=backref('pid_setpoints',
                                            remote_side="ControllerBlock.id"))

    actuator_id = Column(Integer, ForeignKey(ControllerBlock.id), nullable=True)
    actuator = relationship('ControllerBlock',
                            primaryjoin="ControllerBlock.id == PID.actuator_id",
                            foreign_keys=[actuator_id],
                            backref=backref('pid_actuators',
                                            remote_side="ControllerBlock.id"))

    input_id = Column(Integer, ForeignKey(ControllerBlock.id), nullable=True)
    input = ControllerDataField(relationship('ControllerBlock',
                                             foreign_keys=[input_id],
                                             primaryjoin="ControllerBlock.id == PID.input_id",
                                             backref=backref('pid_inputs',
                                                             remote_side="ControllerBlock.id")), writable=True)

    kp = ControllerDataField(Float, writable=True)


    def __repr__(self):
        return '<PID {0}:{1}>'.format(self.profile, self.object_id)



import coloredlogs
import logging

from brewpi_service.datasync import DataSyncherServer

coloredlogs.install(level='DEBUG')

from circuits import Debugger, Component, handler

from brewpi_service import app

LOGGER = logging.getLogger(__name__)

class RestWebServer(Component):
    def __init__(self):
        super(RestWebServer, self).__init__()
        self.app = app

    @handler("started")
    def started(self, *args):
        LOGGER.info("Running Web server...")
        self.app.run()

class Service(Component):
    def init(self):
        self._components = (DataSyncherServer().register(self),
                            RestWebServer().start(process=False, link=self))

service = Service()

Debugger().register(service)

service.run()



from brewpi_service.database import db_session
from brewpi_service.admin import admin, ModelView

from .models import ClockDevice


class ClockDeviceAdminView(ModelView):
    column_hide_backrefs = False
    column_exclude_list = ('type',)
    # column_filters = ('controller',)


admin.add_view(ClockDeviceAdminView(ClockDevice, db_session))



from flask import jsonify

from brewpi_service import app

from .core import plugin_manager, get_plugin
from .schemas import plugin_schema, plugins_schema


@app.route('/system/plugins/')
def plugins():
    """
    List Plugins
    """
    all_plugins = []
    for name, plugin in plugin_manager.all_plugins.items():
        all_plugins.append(plugin)

    result = plugins_schema.dump(all_plugins)
    return jsonify(result.data)


@app.route('/system/plugins/<id>')
def plugin_detail(id):
    """
    Detail a given Plugin
    """
    plugin = get_plugin(id)
    return plugin_schema.jsonify(plugin)



from zope.interface import Interface, Attribute

class ISensor(Interface):
    """
    A mono-value sensor
    """
    sensor_value  = Attribute("""Actual value of this sensor""")


class ISensorSetpointPair(Interface):
    """
    A compound object of a setpoint and a sensor value
    """
    setpoint = Attribute("""The setpoint object""")
    sensor = Attribute("""The sensor object""")


class ISetpoint(Interface):
    """
    A Setpoint for a PID
    """
    value = Attribute("""Actual value of the setpoint""")


class IPID(Interface):
    """
    A proportional–integral–derivative controller (PID controller) is a control
    loop feedback mechanism (controller) commonly used in industrial control
    systems. A PID controller continuously calculates an error value e(t) as
    the difference between a desired setpoint and a measured process variable
    and applies a correction based on proportional, integral, and derivative
    terms (sometimes denoted P, I, and D respectively) which give their name to
    the controller type. -- wikipedia
    """
    enabled = Attribute("""If this PID is enabled""")
    kp = Attribute("""Proportional gain""")
    ti = Attribute("""Integral time constant""")
    td = Attribute("""Derivative time constant""")

    p = Attribute("""Proportional term value, read-only""")
    i = Attribute("""Integral term value, read-only""")
    d = Attribute("""Derivative term value, read-only""")

    actuator_is_negative = Attribute("""If the actuator is negative""")
    input_error = Attribute("""Input error, read-only""")

    set_point = Attribute("""The setpoint object associated with this PID""")
    input_sensor = Attribute("""The Input Sensor object associated with this PID""")
    output_actuator = Attribute("""The Output Actuator object associated with this PID""")



from marshmallow import fields
from brewpi_service import ma

from brewpi_service.controller.schemas import (
    ControllerBlockSchema,
)

from brewpi_service.controller.fields import ControllerData

from .models import (
    TemperatureSensor,
    PID,
    SensorSetpointPair,
    SetpointSimple
)

from .interfaces import (
    ISensor,
    ISetpoint,
    ISensorSetpointPair,
    IPID
)

from zope.interface import implementer


@implementer(ISetpoint)
class SetpointSimpleSchema(ControllerBlockSchema):
    """
    A Simple setpoint
    """
    class Meta:
        model = SetpointSimple
        fields = ControllerBlockSchema.Meta.fields + ('value',)

    value = ControllerData(attribute='value')

@implementer(ISensorSetpointPair)
class SensorSetpointPairSchema(ControllerBlockSchema):
    """
    A simple pair of a Sensor and a Setpoint
    """
    class Meta:
        model = SensorSetpointPair
        fields = ControllerBlockSchema.Meta.fields


@implementer(ISensor)
class TemperatureSensorSchema(ControllerBlockSchema):
    """
    Serialization schema for the TemperatureSensor
    """
    class Meta:
        model = TemperatureSensor
        fields = ControllerBlockSchema.Meta.fields + ('value', 'url')

    value = ControllerData(attribute='value')


@implementer(IPID)
class PIDSchema(ControllerBlockSchema):
    """
    Serialization schema for the PID Loop algorithm
    """
    class Meta:
        model = PID
        fields = ControllerBlockSchema.Meta.fields + ('url', 'kp')

    kp = ControllerData(attribute='kp')

    # actuator = ma.AbsoluteUrlFor('controllerblockdetail', id='<actuator_id>')
    # setpoint = ma.HyperlinkRelated('controllerblockdetail')
    # input = ma.HyperlinkRelated('controllerblockdetail')



#!/usr/bin/env python
import logging

import click

from brewpi_service import app
from brewpi_service.plugins.core import plugin_manager

LOGGER = logging.getLogger(__name__)

@app.cli.command()
def list_routes():
    from urllib.parse import unquote
    from flask import url_for
    output = []
    for rule in app.url_map.iter_rules():

        options = {}
        for arg in rule.arguments:
            options[arg] = "[{0}]".format(arg)

        methods = ','.join(rule.methods)
        url = url_for(rule.endpoint, **options)
        line = unquote("{:50s} {:20s} {}".format(rule.endpoint, methods, url))
        output.append(line)

    for line in sorted(output):
        click.echo(line)


@app.cli.command()
def list_plugins():
    for name, plugin in plugin_manager.all_plugins.items():
        print("{0} [{1}] - {2} (from: {3})".format(name,
                                             '✓' if plugin.enabled else '✗',
                                             plugin.description,
                                             plugin.path))

@app.cli.command()
def initdb():
    from brewpi_service.database import init_db
    click.echo("Initializing db...")
    init_db()


@app.cli.command()
def workers():
    from brewpi_service import app, create_worker_app
    create_worker_app(app)



"""
Generic superclass for aiohttp handlers with startup/cleanup hooks.
"""

import asyncio
from abc import ABC, abstractmethod

from aiohttp import web


class ServiceHandler(ABC):

    def __init__(self, app: web.Application=None):
        if app:
            app.on_startup.append(self._startup)
            app.on_cleanup.append(self._cleanup)

    async def _startup(self, app: web.Application):
        await self.start(app.loop)

    async def _cleanup(self, app: web.Application):
        await self.close()

    @abstractmethod
    async def start(loop: asyncio.BaseEventLoop):
        pass  # pragma: no cover

    @abstractmethod
    async def close():
        pass  # pragma: no cover



import logging

# from brewpi.controlbox.codecs.time import ScaledTime

from brewpi_service.plugins.core import BrewPiServicePlugin

__plugin__ = "ScaledTimeDevicePlugin"

LOGGER = logging.getLogger(__name__)


class ScaledTimeDevicePlugin(BrewPiServicePlugin):
    """
    A Simple plugin that adds support for a Clock Device
    """

    def setup(self):
        from .models import ClockDevice # noqa

    def install(self):
        # Admin
        from . import admin # noqa

        # Data Syncher
        # from brewpi_service.datasync.controlbox import BrewpiEvents
        # from .synchers import ScaledTimeSyncher

        # BrewpiEvents.handlers[ScaledTime] = ScaledTimeSyncher()

        # REST Api
        from .rest import ClockDevice
        from .schemas import ClockSchema
        from brewpi_service.controller.schemas import ControllerBlockDisambiguator
        ControllerBlockDisambiguator.class_to_schema[ClockDevice.__name__] = ClockSchema
        # print(ControllerDeviceDisambiguator.class_to_schema)
        # api.add_resource(ClockDeviceResource, '/clock/<int:id>', endpoint='clockdevice_detail')



"""
Tests functionality offered by brewblox_service.simulator
"""

import pytest
from brewblox_service import simulator

TESTED = simulator.__name__


@pytest.fixture
def sim_config():
    return [
        {
            'identifier': '1.2.3',
            'name': 'beer sensor 1',
            'type_id': 6,
            'settings': {
                'address': '0x2823456512',
                'offset': 0.12
            },
            'state': {
                'value': 20.0,
                'connect': False
            }
        },
        {
            'identifier': '2.3.4',
            'name': 'beer sensor 1',
            'type_id': 6,
            'settings': {
                'address': '0x2823456512',
                'offset': 0.12
            },
            'state': {
                'value': 20.0,
                'connect': True
            }
        }
    ]


@pytest.fixture
async def app(app):
    """App with simulator functions added"""
    simulator.setup(app)
    return app


async def test_config(client, sim_config):
    res = await client.get('/config')
    assert res.status == 200
    assert await res.json() == {}

    # no args supplied
    res = await client.post('/config')
    assert res.status == 500

    # actual config
    res = await client.post('/config', json=sim_config)
    assert res.status == 200

    # now retrieve
    res = await client.get('/config')
    assert res.status == 200
    assert await res.json() == sim_config


async def test_values(client, sim_config):
    res = await client.post('/config', json=sim_config)
    assert res.status == 200

    res = await client.get('/values')
    assert res.status == 200
    content = await res.json()
    assert len(content) == 2
    assert content[0]['identifier'] in ['1.2.3', '2.3.4']

    res = await client.get('/values/1.2.3')
    assert res.status == 200
    content = await res.json()
    assert content['identifier'] == '1.2.3'

    res = await client.get('/values/oilwell')
    assert res.status == 200
    assert await res.json() == dict()



"""
Offers event publishing and subscription for service implementations.

Example use:

    events.setup(app)

    async def on_message(subscription, key, message):
        logging.info(f'Message from {subscription}: {key} = {message} ({type(message)})')

    listener = events.get_listener(app)
    listener.subscribe('brewblox', 'controller', on_message=on_message)
    listener.subscribe('brewblox', 'controller.*', on_message=on_message)
    listener.subscribe('brewblox', 'controller.#', on_message=on_message)
"""

import asyncio
import json
import logging
from concurrent.futures import CancelledError, TimeoutError
from typing import Callable, Type, Union, List
from datetime import timedelta

from aiohttp import web
from brewblox_service.handler import ServiceHandler

import aioamqp

LOGGER = logging.getLogger(__name__)
routes = web.RouteTableDef()

EVENT_CALLBACK_ = Callable[['EventSubscription', str, Union[dict, str]], None]
ExchangeType_ = Type[str]

LISTENER_KEY = 'events.listener'
PUBLISHER_KEY = 'events.publisher'
EVENTBUS_HOST = 'localhost'
EVENTBUS_PORT = 5672
RECONNECT_INTERVAL = timedelta(seconds=1)
PENDING_WAIT_TIMEOUT = timedelta(seconds=5)


def setup(app: Type[web.Application]):
    app[LISTENER_KEY] = EventListener(app)
    app[PUBLISHER_KEY] = EventPublisher(app)
    app.router.add_routes(routes)


def get_listener(app: Type[web.Application]) -> 'EventListener':
    return app[LISTENER_KEY]


def get_publisher(app: Type[web.Application]) -> 'EventPublisher':
    return app[PUBLISHER_KEY]


##############################################################################
# Incoming events
##############################################################################

async def _default_on_message(sub: 'EventSubscription', key: str, message: Union[dict, str]):
    LOGGER.info(f'Unhandled event: subscription={sub}, key={key}, message={message}')


class EventSubscription():
    def __init__(self,
                 exchange_name: str,
                 routing: str,
                 exchange_type: ExchangeType_='topic',
                 on_message: EVENT_CALLBACK_=None):
        self._routing = routing
        self._exchange_name = exchange_name
        self._exchange_type = exchange_type
        self.on_message = on_message

    def __str__(self):
        return f'<{self._routing} @ {self._exchange_name}>'

    @property
    def on_message(self) -> EVENT_CALLBACK_:
        return self._on_message

    @on_message.setter
    def on_message(self, f: EVENT_CALLBACK_):
        self._on_message = f if f else _default_on_message

    async def declare_on_remote(self, channel: aioamqp.channel.Channel):
        LOGGER.info(f'Declaring event bus subscription {self} on {channel}')

        await channel.exchange_declare(
            exchange_name=self._exchange_name,
            type_name=self._exchange_type,
            auto_delete=True
        )

        queue_info = await channel.queue_declare(exclusive=True)
        queue_name = queue_info['queue']

        await channel.queue_bind(
            queue_name=queue_name,
            exchange_name=self._exchange_name,
            routing_key=self._routing
        )

        await channel.basic_consume(
            callback=self._relay,
            queue_name=queue_name
        )

    async def _relay(self,
                     channel: aioamqp.channel.Channel,
                     body: str,
                     envelope: aioamqp.envelope.Envelope,
                     properties: aioamqp.properties.Properties):
        """Relays incoming messages between the queue and the user callback"""
        try:
            await self.on_message(self, envelope.routing_key, json.loads(body))
        except Exception as ex:
            LOGGER.error(f'Exception relaying message in {self}: {ex}')


class EventListener(ServiceHandler):
    def __init__(self, app: web.Application=None, host: str=EVENTBUS_HOST):
        super().__init__(app)

        self._host: str = host

        # Asyncio queues need a context loop
        # We'll initialize self._pending when we have one
        self._pending_pre_async: List[EventSubscription] = []
        self._pending: asyncio.Queue = None
        self._subscriptions: List[EventSubscription] = []

        self._loop: asyncio.BaseEventLoop = None
        self._task: asyncio.Task = None

    def __str__(self):
        return f'<{type(self).__name__} {self._host}>'

    async def start(self, loop: asyncio.BaseEventLoop):
        # Initialize the async queue now we know which loop we're using
        self._pending = asyncio.Queue(loop=loop)

        # Transfer all subscriptions that were made before the event loop started
        [self._pending.put_nowait(s) for s in self._pending_pre_async]

        # We won't be needing this anymore
        self._pending_pre_async = None

        self._task = loop.create_task(self._listen())

    async def close(self):
        LOGGER.info(f'Closing {self}')

        try:
            self._task.cancel()
            await self._task
        except Exception:
            pass
        finally:
            self._task = None

    def subscribe(self,
                  exchange_name: str,
                  routing: str,
                  exchange_type: ExchangeType_='topic',
                  on_message: EVENT_CALLBACK_=None):
        """Adds a new event subscription to the listener.

        Actual queue declaration to the remote message server is done when connected.
        If the listener is not currently connected, it defers declaration.

        on_message(queue, key, message) can be specified.
        If the event was in JSON, message is a dict. Otherwise it is a string

        If on_message() is not set, it will default to logging the message.
        """
        sub = EventSubscription(
            exchange_name,
            routing,
            exchange_type,
            on_message=on_message
        )

        if self._pending is not None:
            self._pending.put_nowait(sub)
        else:
            self._pending_pre_async.append(sub)
            LOGGER.info(f'Deferred event bus subscription: [{sub}]')

        return sub

    async def _listen(self):
        while True:
            try:
                transport, protocol = await aioamqp.connect(
                    host=self._host,
                    loop=self._loop
                )

                channel = await protocol.channel()

                LOGGER.info(f'Connected {self}')

                # Declare all current subscriptions if reconnecting
                [await sub.declare_on_remote(channel) for sub in self._subscriptions]

                while True:
                    subscription = None

                    try:
                        await protocol.ensure_open()
                        subscription = await asyncio.wait_for(
                            self._pending.get(),
                            timeout=PENDING_WAIT_TIMEOUT.seconds
                        )

                    except TimeoutError:
                        # Timeout ensures that connection state is checked at least once per timeout
                        continue

                    try:
                        await protocol.ensure_open()
                        await subscription.declare_on_remote(channel)
                        self._subscriptions.append(subscription)

                    except Exception:
                        # Put subscription back in queue
                        # We'll declare it after reconnect
                        self._pending.put_nowait(subscription)
                        raise

            except CancelledError:
                LOGGER.info(f'Cancelled {self}')
                break

            except ConnectionRefusedError:
                LOGGER.debug(f'Connection refused in {self}')
                await asyncio.sleep(RECONNECT_INTERVAL.seconds)
                continue

            except aioamqp.exceptions.AioamqpException as ex:
                LOGGER.warn(f'Connection error in {self}: {type(ex)}:{ex}')
                await asyncio.sleep(RECONNECT_INTERVAL.seconds)
                continue

            except Exception as ex:
                LOGGER.error(f'Critical error in {self}: {type(ex)}:{ex}', exc_info=True)
                break

            finally:
                try:
                    await protocol.close()
                    transport.close()
                except Exception:
                    pass


##############################################################################
# Outgoing events
##############################################################################

class EventPublisher(ServiceHandler):
    def __init__(self, app: web.Application=None, host: str=EVENTBUS_HOST):
        super().__init__(app)

        self._loop: asyncio.BaseEventLoop = None
        self._host: str = host
        self._reset()

    @property
    def connected(self):
        return self._transport and self._protocol and self._channel

    def __str__(self):
        return f'<{type(self).__name__} {self._host}>'

    async def start(self, loop: asyncio.BaseEventLoop):
        self._loop = loop

    def _reset(self):
        self._transport = None
        self._protocol: aioamqp.AmqpProtocol = None
        self._channel: aioamqp.channel.Channel = None

    async def close(self):
        LOGGER.info(f'Closing {self}')

        try:
            await self._protocol.close()
            self._transport.close()
        except Exception:
            pass
        finally:
            self._reset()

    async def _ensure_channel(self):
        # Lazy connect
        if not self.connected:
            self._transport, self._protocol = await aioamqp.connect(
                host=self._host,
                loop=self._loop
            )
            self._channel = await self._protocol.channel()

        # Assert connection
        try:
            await self._protocol.ensure_open()
        except aioamqp.exceptions.AioamqpException:
            await self.close()
            raise

    async def publish(self,
                      exchange: str,
                      routing: str,
                      message: Union[str, dict]='',
                      exchange_type: ExchangeType_='topic'):
        try:
            await self._ensure_channel()
        except Exception:
            # If server has restarted since our last attempt, ensure channel will fail (old connection invalid)
            # Retry once to check whether a new connection can be made
            await self._ensure_channel()

        # json.dumps() also correctly handles strings
        data = json.dumps(message).encode()

        await self._channel.exchange_declare(
            exchange_name=exchange,
            type_name=exchange_type,
            auto_delete=True
        )

        await self._channel.basic_publish(
            payload=data,
            exchange_name=exchange,
            routing_key=routing
        )


##############################################################################
# REST endpoints
##############################################################################

@routes.post('/_debug/publish')
async def post_publish(request):
    """
    ---
    tags:
    - Events
    summary: Publish event.
    description: Publish a new event message to the event bus.
    operationId: events.publish
    produces:
    - text/plain
    parameters:
    -
        in: body
        name: body
        description: Event message
        required: true
        schema:
            type: object
            properties:
                exchange:
                    type: string
                routing:
                    type: string
                message:
                    type: object
    """
    args = await request.json()
    try:
        await get_publisher(request.app).publish(
            args['exchange'],
            args['routing'],
            args['message']
        )
        return web.Response()

    except Exception as ex:
        LOGGER.warn(f'Unable to publish {args}: {ex}')
        return web.Response(body='Event bus connection refused', status=500)


@routes.post('/_debug/subscribe')
async def post_subscribe(request):
    """
    ---
    tags:
    - Events
    summary: Subscribe to events.
    operationId: events.subscribe
    produces:
    - text/plain
    parameters:
    -
        in: body
        name: body
        description: Event message
        required: true
        schema:
            type: object
            properties:
                exchange:
                    type: string
                routing:
                    type: string
    """
    args = await request.json()
    get_listener(request.app).subscribe(
        args['exchange'],
        args['routing']
    )
    return web.Response()



from brewpi_service.database import db_session
from brewpi_service.admin import admin, ModelView

from .models import LoggedDeviceConfiguration


class LoggedDeviceConfigurationAdminView(ModelView):
    column_hide_backrefs = False


admin.add_view(LoggedDeviceConfigurationAdminView(LoggedDeviceConfiguration,
                                                  db_session))



import coloredlogs

from flask import Flask
from flask_marshmallow import Marshmallow
from flask_cors import CORS
from apispec import APISpec

coloredlogs.install(level='DEBUG')

# FIXME Should be in a configuration file

CONFIG = {
    'APISPEC_SPEC': APISpec(
        title='BrewPi Service',
        version='0.1',
        plugins=['apispec.ext.marshmallow']
    ),
    'APISPEC_SWAGGER_URL': '/specs.json',
    'APISPEC_SWAGGER_UI_URL': None,
    "SECRET_KEY": 'A0Zr98j/3yX R~XHH!jmN]LWX/,?RT',
    "DATABASE_URI": 'postgresql+psycopg2://bbox_service:bbox_service@localhost/bbox_service',
    "RQ_REDIS_URL": "redis://localhost:6379",
    "SSE_REDIS_URL": "redis://localhost"
}

app = Flask("brewpi_service")
app.config.update(CONFIG)

from .tasks import rq
rq.init_app(app)

ma = Marshmallow(app)


from . import commands

from . import controller
from .controller.models import ControllerBlock
from .database import db_session, load_models

from .plugins.rest import *
from . import rest

from .controller import rest


def create_front_app():
    cors = CORS(app)

    from . import sse

    from .plugins.core import plugin_manager
    plugin_manager.init_app(app=app, plugin_folder="plugins_dir")


    from .rest import api_v1, specs as api_specs
    api_specs.init_app(app)
    api_v1.init_app(app)

    # Now, let plugins register things
    plugin_manager.install_plugins()


    # Init application components
    from .admin import admin
    admin.init_app(app)

    return app

create_front_app()






from werkzeug.exceptions import NotFound

from flask_apispec import MethodResource
from flask_apispec import marshal_with

from brewpi_service.rest import api_v1
from brewpi_service.datasync.backends.cache import available_blocks_cache

from .models import Controller, ControllerBlock, ControllerProfile
from .schemas import (
    ControllerSchema,
    ControllerBlockSchema,
    ControllerAvailableBlockSchema,
    ControllerProfileSchema
)


@marshal_with(ControllerSchema(many=True))
class ControllerList(MethodResource):
    def get(self, **kwargs):
        return Controller.query.all()

api_v1.register('/controllers/', ControllerList)


@marshal_with(ControllerSchema)
class ControllerDetail(MethodResource):
    def get(self, id):
        return Controller.query.get(id)

api_v1.register('/controllers/<id>/', ControllerDetail)


@marshal_with(ControllerAvailableBlockSchema(many=True))
class ControllerAvailableBlockList(MethodResource):
    """
    Available Blocks from the Controller
    """
    def get(self, controller_id):
        controller = Controller.query.get(controller_id)
        if controller is None:
            raise NotFound()
        return available_blocks_cache.get_all_for(controller)

api_v1.register('/controllers/<controller_id>/available_blocks/', ControllerAvailableBlockList)


@marshal_with(ControllerProfileSchema)
class ControllerProfileDetail(MethodResource):
    def get(self, id):
        return ControllerProfile.query.get(id)

api_v1.register('/controller_profiles/<id>/', ControllerProfileDetail)


@marshal_with(ControllerBlockSchema(many=True))
class ControllerBlockList(MethodResource):
    def get(self):
        return ControllerBlock.query.all()

api_v1.register('/controller_blocks/', ControllerBlockList)


@marshal_with(ControllerBlockSchema)
class ControllerBlockDetail(MethodResource):
    def get(self, id):
        return ControllerBlock.query.get(id)

api_v1.register('/controller_blocks/<id>/', ControllerBlockDetail)



import datetime

from sqlalchemy import (
    Column, Integer, String, Boolean, ForeignKey,
    UniqueConstraint, DateTime
)
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr

from ..database import Base, db_session


class TimestampMixin:
    """
    Timestamp mixin for keeping track of creation time and update of objects
    """
    created_at = Column(DateTime, default=datetime.datetime.utcnow, nullable=False)
    updated_at = Column(DateTime, default=datetime.datetime.utcnow, onupdate=datetime.datetime.utcnow, nullable=False)


class BaseObject(Base, TimestampMixin):
    """
    Basic object either living inside or outside the electronic world
    """
    __abstract__ = True

    id = Column(Integer, primary_key=True)
    type = Column(String(50))


class Device(BaseObject):
    """
    An device not living on a controller, i.e. a manually triggered device.
    """
    __tablename__ = "device"
    __mapper_args__ = {
        'polymorphic_identity': "device",
        'polymorphic_on': 'type',
        'with_polymorphic': '*'
    }


class ControllerBlock(BaseObject):
    """
    Any block (i.e. object) living in the controller
    """
    __tablename__ = "controller_block"
    __mapper_args__ = {
        'polymorphic_identity': "controller_block",
        'polymorphic_on': 'type',
        'with_polymorphic': '*'
    }

    def __setattr__(self, name, value):
        if name in self.__class__._controller_data_fields:
            attribute = getattr(self, name)
            if type(value) is tuple:
                attribute.set_actual_value(value[0])
                attribute.request_value(value[1])
            else:
                attribute.request_value(value)

            return value
        else:
            return super().__setattr__(name, value)


    def get_dirty_fields(self):
        dirty_fields = []
        for field in self.__class__._controller_data_fields:
            if getattr(self, field)[2] == True:
                dirty_fields.append(field)

        return dirty_fields


    def get_actual_value_of_field(self, field_name):
        if name in self.__class__._controller_data_fields:
            attribute = getattr(self, name)
            attribute.request_value(value)


    is_static = Column(Boolean, default=False, nullable=False, doc="whether the object is hardcoded on the controller or user-creatable")

    profile_id = Column(Integer, ForeignKey('controller_profile.id'), nullable=True)
    profile = relationship("models.ControllerProfile")

    object_id = Column(Integer, nullable=False)
    name = Column(String, nullable=True)

    __table_args__ = (
        UniqueConstraint('profile_id', 'object_id', name='_controller_object_uc'),
    )


class Controller(Base):
    """
    A Hardware Controller that holds sensors and actuators
    """
    __tablename__ = 'controller'

    id = Column(Integer, primary_key=True)
    name = Column(String(64), index=True, unique=True)
    uri = Column(String(128), index=True, unique=True)
    description = Column(String(128))
    connected = Column(Boolean)

    profile_id = Column(Integer, ForeignKey('controller_profile.id'))
    profile = relationship("ControllerProfile", back_populates="controllers")

    def __repr__(self):
        return '<Controller {0}@{1} using profile "{2}">'.format(self.name, self.uri, self.profile)


class ControllerProfile(Base):
    """
    A Profile for a Controller
    """
    __tablename__ = 'controller_profile'

    id = Column(Integer, primary_key=True)
    name = Column(String(64), index=True, unique=True)

    controllers = relationship("Controller", back_populates="profile")

    blocks = relationship("ControllerBlock", back_populates="profile")

    def get_block_by_name(self, name):
        return ControllerBlock.query.with_polymorphic('*').filter(ControllerBlock.name==name).one()

    def __repr__(self):
        return self.name



# from celery import Celery
from flask_rq2 import RQ

rq = RQ()

# from .datasync.controlbox import ControlboxSyncher
# from .datasync.brewpi_legacy import LegacySyncher
#    db_syncher = DatabaseControllerSyncher()




from brewpi_service import ma


class PluginSchema(ma.Schema):
    id = ma.String(attribute="identifier")
    url = ma.AbsoluteUrlFor('plugin_detail', id='<identifier>')
    name = ma.String()
    description = ma.String()
    author = ma.String()
    license = ma.String()
    version = ma.String()
    website = ma.String()
    enabled = ma.Boolean()


plugin_schema = PluginSchema()
plugins_schema = PluginSchema(many=True)



from brewpi_service.rest import (
    marshal_with,
    MethodResource,
    api_v1
)

from .models import ClockDevice
from .schemas import ClockSchema


@marshal_with(ClockSchema(many=True))
class ClockList(MethodResource):
    """
    List Clocks
    """
    def get(self, **kwargs):
        return ClockDevice.query.all()

api_v1.register('/clocks/', ClockList)


@marshal_with(ClockSchema)
class ClockDetail(MethodResource):
    """
    Detail a given Clock
    """
    def get(self, id, **kwargs):
        return ClockDevice.query.get(id)

api_v1.register('/clocks/<id>/', ClockDetail)



from abc import abstractmethod

from circuits import Component

class AbstractControllerSyncherBackend:
    """
    A Syncher Backend implements a way to sync a Hardware Device (Controller)
    with the service.
    """
    pass

class AbstractForwarder:
    """
    Forwards synched data to some other systems so they can sync at their turn.
    """
    @abstractmethod
    def on_controller_appeared(self, event):
        raise NotImplementedError

    @abstractmethod
    def on_controller_disappeared(self, event):
        raise NotImplementedError


class AbstractBackstoreSyncher:
    """
    Implements a syncher that takes care of receiving events from `Backends`
    and persist data into a backing store.
    """
    @abstractmethod
    def on_controller_appeared(self, event):
        raise NotImplementedError

    @abstractmethod
    def on_controller_disappeared(self, event):
        raise NotImplementedError


class AbstractDeviceSyncher:
    @abstractmethod
    def update(self, controller, state, event):
        raise NotImplementedError



"""
Master file for pytest fixtures.
Any fixtures declared here are available to all test functions in this directory.
"""


import logging

import pytest

from brewblox_service import service


@pytest.fixture(scope='session', autouse=True)
def log_enabled():
    """Sets log level to DEBUG for all test functions.
    Allows all logged messages to be captured during pytest runs"""
    logging.getLogger().setLevel(logging.DEBUG)


@pytest.fixture
def app_config() -> dict:
    return {
        'name': 'test_app',
        'host': '0.0.0.0',
        'port': 1234,
        'debug': False,
        'output': None,
    }


@pytest.fixture
def sys_args(app_config) -> list:
    return [
        'app_name',
        '--name', app_config['name'],
        '--host', app_config['host'],
        '--port', str(app_config['port']),
    ]


@pytest.fixture
def app(sys_args):
    app = service.create_app('default', raw_args=sys_args[1:])
    return app


@pytest.fixture
def client(app, aiohttp_client, loop):
    """Allows patching the app or aiohttp_client before yielding it.

    Any tests wishing to add custom behavior to app can override the fixture
    """
    return loop.run_until_complete(aiohttp_client(app))



from setuptools import setup, find_packages


setup(
    name='brewblox-service',
    use_scm_version={'local_scheme': lambda v: ''},
    long_description=open('README.md').read(),
    url='https://github.com/BrewBlox/brewblox-service',
    author='BrewPi',
    author_email='development@brewpi.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python :: 3.6',
        'Intended Audience :: End Users/Desktop',
        'Topic :: System :: Hardware',
    ],
    keywords='brewing brewpi brewblox embedded',
    packages=find_packages(exclude=['test']),
    install_requires=[
        'pprint',
        'dpath',
        'aiohttp',
        'aiodns',
        'cchardet',
        'aiohttp-cors',
        'aiohttp-swagger',
        'aioamqp',
    ],
    python_requires='>=3.6',
    extras_require={'dev': ['tox']},
    setup_requires=['setuptools_scm']
)



from sqlalchemy import (
    Column, Integer, ForeignKey
)

from brewpi_service.controller.state import ControllerDataField
from brewpi_service.controller.models import ControllerBlock


class ClockDevice(ControllerBlock):
    """
    A clock Sensor
    """
    __tablename__ = 'clock_device'

    __mapper_args__ = {
        'polymorphic_identity': "clock_device"
    }
    id = Column(Integer, ForeignKey('controller_block.id'), primary_key=True)

    scale = ControllerDataField(Integer, writable=True)
    time = ControllerDataField(Integer)

    def __repr__(self):
        return '<Clock {0} -> {1} (x{2})>'.format(self.id, self.time, self.scale)



import logging
from sqlalchemy.sql import exists
from sqlalchemy.orm.exc import NoResultFound

from ..database import db_session

from .models import ControllerProfile

LOGGER = logging.getLogger(__name__)


class ControllerProfileManager:
    def get(self, name):
        return db_session.query(ControllerProfile).filter(ControllerProfile.name==name).one()

    def exists(self, name):
        return db_session.query(ControllerProfile).filter(exists().where(ControllerProfile.name==name)).first()

    def create(self, name):
        new_profile = ControllerProfile(name=name)
        db_session.add(new_profile)
        db_session.commit()
        LOGGER.debug("Created new profile {0}".format(name))

        return new_profile



ControllerProfileManager = ControllerProfileManager()




"""
Tests functionality offered by brewblox_service.events
"""

import asyncio
import json
from datetime import timedelta
from unittest.mock import Mock, call

import aioamqp
import pytest
from asynctest import CoroutineMock
from brewblox_service import events

TESTED = events.__name__


@pytest.fixture
def channel_mock():
    m = Mock()
    m.exchange_declare = CoroutineMock()
    m.queue_declare = CoroutineMock(return_value={'queue': 'queue_name'})
    m.queue_bind = CoroutineMock()
    m.basic_consume = CoroutineMock()
    m.basic_publish = CoroutineMock()
    return m


@pytest.fixture
def protocol_mock(channel_mock):
    m = Mock()
    m.channel = CoroutineMock(return_value=channel_mock)
    m.ensure_open = CoroutineMock()
    m.close = CoroutineMock()
    return m


@pytest.fixture
def transport_mock():
    return Mock()


@pytest.fixture
def mocked_connect(mocker, protocol_mock, transport_mock):
    # connect()
    conn_mock_func = CoroutineMock(spec=aioamqp.connect)
    conn_mock_func.return_value = (transport_mock, protocol_mock)

    mocker.patch.object(events.aioamqp, 'connect', conn_mock_func)

    return conn_mock_func


@pytest.fixture
async def app(app, mocker, loop, mocked_connect):
    """App with events enabled"""
    events.setup(app)
    mocker.patch(TESTED + '.PENDING_WAIT_TIMEOUT', timedelta(microseconds=10))
    mocker.patch(TESTED + '.RECONNECT_INTERVAL', timedelta(microseconds=10))
    return app


async def test_setup(app, client):
    assert events.get_listener(app)
    assert events.get_publisher(app)


async def test_subscribe_callbacks(app, client, mocker):
    cb = Mock()
    content = dict(key='var')

    chan_mock = Mock()
    body = json.dumps(content)
    envelope_mock = Mock()
    envelope_mock.routing_key = 'message_key'
    properties_mock = Mock()

    sub = events.get_listener(app).subscribe(
        'brewblox', 'router', on_message=cb)

    await sub._relay(chan_mock, body, envelope_mock, properties_mock)
    cb.assert_called_once_with(sub, 'message_key', content)

    cb2 = Mock()
    sub.on_message = cb2

    await sub._relay(chan_mock, body, envelope_mock, properties_mock)
    cb2.assert_called_once_with(sub, 'message_key', content)
    assert cb.call_count == 1

    # Shouldn't blow up
    cb2.side_effect = ValueError
    await sub._relay(chan_mock, body, envelope_mock, properties_mock)

    # Should be tolerated
    sub.on_message = None
    await sub._relay(chan_mock, body, envelope_mock, properties_mock)


async def test_offline_listener(app, mocker):
    # We're not importing the client fixture, so the app is not running
    # Expected behaviour is for the listener to add functions to the app hooks
    listener = events.EventListener(app)

    assert listener._startup in app.on_startup
    assert listener._cleanup in app.on_cleanup

    # Subscriptions will not be declared yet - they're waiting for startup
    sub = listener.subscribe('exchange', 'routing')
    assert sub in listener._pending_pre_async

    # Can safely be called, but will be a no-op at this time
    await listener.close()


async def test_online_listener(app, client, mocker):
    # The client fixture has started running the app in the event loop
    # We'll need to start explicitly
    with pytest.raises(RuntimeError):
        events.EventListener(app)

    listener = events.EventListener()
    sub = listener.subscribe('exchange', 'routing')

    # No-op, listener is not yet started
    await listener.close()

    assert sub in listener._pending_pre_async
    await listener.start(app.loop)
    assert listener._pending_pre_async is None

    pending_subs = listener._pending.qsize()
    listener.subscribe('exchange', 'routing')
    # We haven't yielded, so the subscription can't have been processed yet
    assert listener._pending.qsize() == pending_subs + 1

    # Safe for repeated calls
    await listener.close()
    await listener.close()


async def test_offline_publisher(app):
    publisher = events.EventPublisher(app)

    assert publisher._startup in app.on_startup
    assert publisher._cleanup in app.on_cleanup

    # with pytest.raises(ConnectionRefusedError):
    await publisher.publish('exchange', 'key', message=dict(key='val'))


async def test_online_publisher(app, client, mocker):
    # Can't setup while running
    with pytest.raises(RuntimeError):
        events.EventPublisher(app)

    publisher = events.EventPublisher()
    await publisher.start(app.loop)

    await publisher.publish('exchange', 'key', message=dict(key='val'))
    await publisher.publish('exchange', 'key', message=dict(key='val'))


async def test_publish_endpoint(app, client, mocker):
    publish_spy = mocker.spy(events.get_publisher(app), 'publish')

    # standard ok
    assert (await client.post('/_debug/publish', json=dict(
        exchange='exchange',
        routing='first',
        message=dict(key='val')
    ))).status == 200

    # string messages supported
    assert (await client.post('/_debug/publish', json=dict(
        exchange='exchange',
        routing='second',
        message='message'
    ))).status == 200

    # return 500 on connection refused
    events.get_publisher(app)._ensure_channel = CoroutineMock(side_effect=ConnectionRefusedError)
    assert (await client.post('/_debug/publish', json=dict(
        exchange='exchange',
        routing='third',
        message='message'
    ))).status == 500

    publish_spy.assert_has_calls([
        call('exchange', 'first', dict(key='val')),
        call('exchange', 'second', 'message'),
        call('exchange', 'third', 'message'),
    ])


async def test_subscribe_endpoint(app, client, channel_mock):
    assert (await client.post('/_debug/subscribe', json=dict(
        exchange='exchange',
        routing='routing.key'
    ))).status == 200

    await asyncio.sleep(0.01)

    channel_mock.queue_bind.assert_called_once_with(
        queue_name='queue_name',
        exchange_name='exchange',
        routing_key='routing.key'
    )


async def test_listener_exceptions(app, client, protocol_mock, channel_mock, transport_mock, mocked_connect):
    listener = events.get_listener(app)

    channel_mock.basic_consume.side_effect = ConnectionRefusedError
    listener.subscribe('exchange', 'routing')

    # Let listener attempt to declare the new subscription a few times
    # In real scenarios we'd expect ensure_open() to also start raising exceptions
    await asyncio.sleep(0.01)
    protocol_mock.ensure_open.side_effect = aioamqp.AmqpClosedConnection
    await asyncio.sleep(0.01)

    # Retrieved subscription, errored on basic_consume(), and put subscription back
    assert channel_mock.basic_consume.call_count >= 1
    assert listener._pending.qsize() == 1
    assert not listener._task.done()

    # Recoverable error
    protocol_mock.ensure_open.side_effect = aioamqp.AmqpClosedConnection
    await asyncio.sleep(0.01)
    assert not listener._task.done()

    # Critical error
    protocol_mock.ensure_open.side_effect = RuntimeError
    await asyncio.sleep(0.01)
    assert listener._task.done()

    # Should be closed every time it was opened
    assert protocol_mock.close.call_count == mocked_connect.call_count
    assert transport_mock.close.call_count == mocked_connect.call_count


async def test_listener_periodic_check(mocker, app, client, loop, protocol_mock):
    # Check that the listener periodically called ensure_open to check
    start_count = protocol_mock.ensure_open.call_count
    await asyncio.sleep(0.1)
    assert protocol_mock.ensure_open.call_count > start_count


async def test_listener_close_error(app, client, loop, mocked_connect):
    mocked_connect.side_effect = ConnectionRefusedError
    listener = events.EventListener()
    await listener.start(loop)
    await asyncio.sleep(0.01)

    # ConnectionRefused is deemed recoverable
    assert not listener._task.done()

    # Closed after unrecoverable error, but application still alive
    mocked_connect.side_effect = RecursionError
    await asyncio.sleep(0.01)
    assert listener._task.done()


async def test_publisher_exceptions(app, client, protocol_mock):
    protocol_mock.ensure_open.side_effect = aioamqp.AmqpClosedConnection

    with pytest.raises(aioamqp.AmqpClosedConnection):
        await events.get_publisher(app).publish('exchange', 'routing', 'message')
